/*
 * Copyright 2012-2017 Red Hat, Inc.
 *
 * This file is part of Thermostat.
 *
 * Thermostat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2, or (at your
 * option) any later version.
 *
 * Thermostat is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Thermostat; see the file COPYING.  If not see
 * <http://www.gnu.org/licenses/>.
 *
 * Linking this code with other modules is making a combined work
 * based on this code.  Thus, the terms and conditions of the GNU
 * General Public License cover the whole combination.
 *
 * As a special exception, the copyright holders of this code give
 * you permission to link this code with independent modules to
 * produce an executable, regardless of the license terms of these
 * independent modules, and to copy and distribute the resulting
 * executable under terms of your choice, provided that you also
 * meet, for each linked independent module, the terms and conditions
 * of the license of that module.  An independent module is a module
 * which is not derived from or based on this code.  If you modify
 * this code, you may extend this exception to your version of the
 * library, but you are not obligated to do so.  If you do not wish
 * to do so, delete this exception statement from your version.
 */

package com.redhat.thermostat.gateway.service.schema;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonObject;

import com.redhat.thermostat.common.yaml.JsonToYaml;
import com.redhat.thermostat.gateway.common.util.LoggingUtil;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;

import java.io.IOException;
import java.io.StringWriter;

import java.util.logging.Logger;

@Path("/")
public class SchemaHttpHandler {

    private static final Logger logger = LoggingUtil.getLogger(SchemaHttpHandler.class);

    private static final Object combineLock = new Object();
    private static JsonObject combinedSwagger = null;
    private static boolean haveTriedToCombine = false;
    private Gson gson = null;
    private Gson prettyGson = null;

    @GET
    @Path("/fullapi-swagger.yaml")
    @Produces({ MediaType.TEXT_PLAIN })
    public Response getCombinedAPI(@Context ServletContext context) throws Exception {
        return getCombinedAPI(context, true, false);
    }

    @GET
    @Path("/fullapi-swagger.json")
    @Produces({ MediaType.APPLICATION_JSON })
    public Response getCombinedAPI(@QueryParam("pretty") @DefaultValue("false") Boolean prettyPrint,
                                   @Context ServletContext context) throws Exception {

        return getCombinedAPI(context, false, prettyPrint);
    }

    @GET
    @Path("/version")
    @Produces({ MediaType.APPLICATION_JSON })
    public Response getVersion() throws Exception {

        final String versionStr = "{ \"version\" : \"0.0.0\" }";
        return Response.status(Response.Status.OK).type(MediaType.APPLICATION_JSON_TYPE).entity(versionStr).build();
    }

    private Response getCombinedAPI(ServletContext context, boolean toYaml, boolean prettyPrint) throws IOException {
        if (!haveTriedToCombine) {
            synchronized (combineLock) {
                if (!haveTriedToCombine) {
                    try {
                        // this is a long running operation
                        combinedSwagger = new SwaggerBuilder().createCombinedSwagger(context);
                    } catch (Exception e) {
                        logger.warning("error creating combined Swagger API: " + e.getLocalizedMessage());
                    }
                    haveTriedToCombine = true;
                }
            }
        }

        if (combinedSwagger != null) {
            final StringWriter out = new StringWriter();
            final MediaType mediaType;
            if (toYaml) {
                new JsonToYaml().jsonToYaml(combinedSwagger, out);
                mediaType = MediaType.TEXT_PLAIN_TYPE;
            } else {
                getGsonInstance(prettyPrint).toJson(combinedSwagger, out);
                mediaType = MediaType.APPLICATION_JSON_TYPE;
            }
            return Response.status(Response.Status.OK).type(mediaType).entity(out.toString()).build();
        } else {
            return Response.status(Response.Status.INTERNAL_SERVER_ERROR).build();
        }
    }

    private Gson getGsonInstance(boolean isPretty) {
        if (isPretty) {
            if (prettyGson == null) {
                final GsonBuilder gsonBuilder = new GsonBuilder();
                gsonBuilder.setPrettyPrinting();
                prettyGson = gsonBuilder.create();
            }
            return prettyGson;
        } else {
            if (gson == null) {
                gson = new Gson();
            }
            return gson;
        }
    }
}
